#include <ros/init.h>
#include <ros/node_handle.h>
#include <ros/param.h>
#include <opencv2/opencv.hpp>
#include <node_msgs/arucoPose.h>
#include <sensor_msgs/Image.h>
#include "ImageLoader.h"
#include "ArucoPos.h"
#include "ArucoCallback.h"
#include "PrintToFile.h"

int main(int argc, char** argv) {

    // Initializing ROS
	ros::init(argc, argv, "node_aruco");

    // Creates a node handle
	ros::NodeHandle nH;
	ROS_INFO("ARUCO: Starting up aruco detection node...");

    // Run the node at 5Hz
	double loopRate;
	ros::param::param<double>("aruco_loop_rate", loopRate, 5.0);
	ROS_INFO("ARUCO: Using loop rate of %f Hz", loopRate);
	
	ros::Rate rate(loopRate);
	
	std::string outputImageTopic;
	ros::param::param<std::string>("output_image", outputImageTopic, "/visnav/aruco_track");
	ros::Publisher outputImagePub = nH.advertise<sensor_msgs::Image>(outputImageTopic, 1);
	
	// Publisher
    ros::Publisher pub = nH.advertise<node_msgs::arucoPose>("/arucoNode/data",1000);
    node_msgs::arucoPose poseMsg;


    //GNSS time_ref
    // Subscribes to the /mavros/time_reference
    ArucoCallback arucoCallback;
    ros::Subscriber time_sub = nH.subscribe("/mavros/time_reference", 10, &ArucoCallback::gnssTimeCallback, &arucoCallback);
    ROS_INFO("ARUCO: Expecting time reference on topic: /mavros/time_reference" );

	//ROS_INFO("ARUCO: Waiting for first time reference");
	while (arucoCallback.hasDt() == false ) {
		ros::spinOnce();
		rate.sleep();
	}
	ros::Duration timeCorrection = arucoCallback.getDtGnssRos();
	ROS_INFO("ARUCO: Delta time reference is set to %i s and %i ns", timeCorrection.sec, timeCorrection.nsec);


    //Image Loader
	std::string imageTopic;
	ros::param::param<std::string>("image_topic", imageTopic, "/camera/image_raw");
    ROS_INFO("ARUCO: Expecting images on topic: %s ", imageTopic.c_str());

	vis_nav::ImageLoader imageLoader(true);
    ros::Subscriber img_sub = nH.subscribe(imageTopic.c_str(),   1,  &vis_nav::ImageLoader::imageCallback  , &imageLoader);

	ROS_INFO("ARUCO: Waiting for first image");
	while (imageLoader.hasImage() == false ) {
		ros::spinOnce();
		rate.sleep();
	}


	ROS_INFO("ARUCO: Imageloader ready!");
	
	int imageWidth  = imageLoader.getImageWidth();
	int imageHeight = imageLoader.getImageHeight();
	
	ROS_INFO("ARUCO: Image width:  %i ", imageWidth);
	ROS_INFO("ARUCO: Image height: %i ", imageHeight);

	ROS_INFO("ARUCO: ARUCO detection node is looking for good parameters");	
	/*
	 * 	Setting up good features to track
	 */
	
	cv::Mat currentImage(imageHeight, imageWidth, CV_8UC1);
	ros::Time imageTimeStamp;

	ArucoPos tagDetection(false,imageLoader.getImage()->header,imageLoader.getImage()->encoding);
	tagDetection.readParametersFromFile();
	tagDetection.setDictionary(cv::aruco::getPredefinedDictionary(cv::aruco::DICT_5X5_50));
	tagDetection.setParameters(cv::aruco::DetectorParameters::create());
	
	ROS_INFO("ARUCO: detection node is starting...");

	//Data from tag search:
	vector<avTag> dataFromAruco;


	while (ros::ok() ) {
	
		ros::spinOnce();

		//Load image
		currentImage = imageLoader.getImageMatrix(CV_8UC1);
		imageTimeStamp = imageLoader.getTimeStamp();
        //ROS_INFO("ARUCO: imageTimeStamp is set to %i s and %i ns", imageTimeStamp.sec, imageTimeStamp.nsec);
		
		dataFromAruco = tagDetection.searchForTag(currentImage);

		tagDetection.printImage(outputImagePub);

		for(size_t i=0;i<dataFromAruco.size();i++)
		{
			poseMsg.header.stamp=imageTimeStamp;

			poseMsg.header.frame_id="/world";

			poseMsg.vehicleId=dataFromAruco.at(i).VehId;

			poseMsg.pose.position.x=dataFromAruco.at(i).tVector[0];
			poseMsg.pose.position.y=dataFromAruco.at(i).tVector[1];
			poseMsg.pose.position.z=dataFromAruco.at(i).tVector[2];

			poseMsg.pose.orientation.x=dataFromAruco.at(i).rVector.x();
			poseMsg.pose.orientation.y=dataFromAruco.at(i).rVector.y();
			poseMsg.pose.orientation.z=dataFromAruco.at(i).rVector.z();
			poseMsg.pose.orientation.w=dataFromAruco.at(i).rVector.w();

			pub.publish(poseMsg);

            // Print to file
            //if(enablePrint) pf.writeData(poseMsg);

		}
		dataFromAruco.clear();

		rate.sleep();

	}

    //pf.closeFile();

	ROS_INFO("ARUCO: Detection node is terminated");
	
	return 0;

}

//0,287558685