//
// Created by line on 28.02.18.
//
// in cmake options: -DCATKIN_DEVEL_PREFIX=../../devel

#include <ros/init.h>
#include <ros/node_handle.h>
#include <ros/param.h>
#include "node_controller/tools/autopilot.h"
#include <node_msgs/navResults.h>
#include <node_msgs/cmd_velocity.h>
#include <node_msgs/cmd_info.h>
#include "node_controller/tools/navigationCallback.h"

int main(int argc, char** argv) {
    // Initializing ROS
    ros::init(argc, argv, "node_controller");

    // Creates a node handle
    ros::NodeHandle nH;
    ROS_INFO("CONTROLLER: Starting up controller node...");

    // Run the node at 20Hz
    double loopRate;
    ros::param::param<double>("controller_loop_rate", loopRate, 10.0);
    ROS_INFO("CONTROLLER: Using loop rate of %f Hz", loopRate);

    ros::Rate rate(loopRate);

    //Callback form the state estimator (Kalman filter)
    NavigationCallback navCallback;
    ros::Subscriber navSub = nH.subscribe("/navigationNode/stateEstimator",10,&NavigationCallback::statesCallback, &navCallback);
    ROS_INFO("CONTROLLER: Expecting estimates states on topic: /navigationNode/stateEstimator" );

    //Callback from UAV_control_system
    ros::Subscriber statusSub = nH.subscribe("/cmd/status",10,&NavigationCallback::cmdStatusCallback, &navCallback);
    ROS_INFO("CONTROLLER: Expecting control status on topic: /cmd/status" );

    // Publisher
    ros::Publisher pub = nH.advertise<node_msgs::cmd_velocity>("/cmd/velocity",1);
    node_msgs::cmd_velocity velMsg;

    ros::Publisher pub2 = nH.advertise<node_msgs::cmd_info>("/cmd/info",1);
    node_msgs::cmd_info infoMsg;


    //Read parameters from the controlParameters.yaml file
    Autopilot ap;
    ap.loadParameterFile();

    Eigen::Vector3d relPos;
    Eigen::Matrix3d relPosCovar;
    Eigen::Vector3d velLandingPad;
    Eigen::Vector3d velCommand;
    bool controlGiven;

    int mode =Autopilot::mode::stop;

    ros::Time prevTime = ros::Time::now();
    double dt;
    while (ros::ok() ) {

        ros::spinOnce();

        //Read state estimates from the Kalman filter
        relPos=navCallback.getStates().rel_position;
        relPosCovar=navCallback.getStates().rel_position_covariance;
        velLandingPad=navCallback.getStates().lp_vel;

        //Read out control given
        controlGiven=navCallback.getStatus().controlGiven;


        //Select mode
        mode=ap.modeSelector(mode,controlGiven,relPos,relPosCovar);

        //Print out active mode
        //std::cout << "Mode: " << mode << std::endl;

        //Get dt
        dt=(ros::Time::now()-prevTime).toSec();
        prevTime=ros::Time::now();

        //Run the controller
        velCommand=ap.controller(mode,relPos,velLandingPad,dt);

        //Send land command and velocity sp
        velMsg.header.stamp=prevTime;
        velMsg.header.frame_id="NED";
        velMsg.do_land=ap.getEnambleLand();
        velMsg.velocity_north=velCommand.x();
        velMsg.velocity_east=velCommand.y();
        velMsg.velocity_down=velCommand.z();
        velMsg.status=mode;
        pub.publish(velMsg);


        //Send info msg
        infoMsg.header.stamp=prevTime;
        infoMsg.do_land=ap.getEnambleLand();
        infoMsg.status=mode;
        infoMsg.hover_height=ap.getHoverHeight();
        pub2.publish(infoMsg);

        rate.sleep();
    }


    return 0;
}