//
// Created by line on 28.02.18.
//

#include <node_controller/tools/autopilot.h>

Autopilot::Autopilot(){
    enableLand_=false;
    dynhoverHeight_=hoverHeight_;
}

int Autopilot::modeSelector(int md, bool control_given, Eigen::Vector3d p_u_t, Eigen::Matrix3d p_u_t_covar) {
    int md_out = md;
    Eigen::Vector3d hoverVectTop(0, 0, -hoverHeight_);
    if(!control_given) md_out=mode::stop;
    else{
        switch(md){
            case mode::stop : {
                md_out = mode::intercept;
                break;
            }

            case mode::intercept: {
                if (posTolIntercept_ > (p_u_t + hoverVectTop).norm()) md_out = mode::hover;        //sphere
                break;
            }

            case mode::hover: {
                //Todo: Add covariance check
                md_out = mode::lower;
                break;
            }

            case mode::lower: {
                if(p_u_t.z() < gainAdjustAlt_) md_out=mode::gainAdjust;
                break;
            }

            case mode::gainAdjust: {
                if(p_u_t.z() < finalStageAlt_) md_out=mode::finalStage;
                break;
            }

            case mode::finalStage:{
                if ((sqrt(pow(p_u_t.x(), 2.0) + pow(p_u_t.y(), 2.0)) < posTolLowerRadius_) &&
                    (p_u_t.z() < posTolLowerHeight_)) { //Cylinder
                    md_out = mode::land;
                }
                break;
            }

            case mode::land: {
                if (uavLanded) md_out = mode::landed;
                break;
            }

            default: {
                md_out = md;
            }
        }
    }


    return md_out;
}

Eigen::Vector3d Autopilot::controller(int md,Eigen::Vector3d p_u_t, Eigen::Vector3d v_n_t, double dt) {
    Eigen::Vector3d out;
    Eigen::Vector3d hoverVectTop(0,0,-hoverHeight_);
    switch (md){
        case mode::stop : {
            enableLand_ = false;
            out = Eigen::Vector3d::Zero(3, 1);
            break;
        }

        case mode::intercept :{}
        case mode::hover : {
            enableLand_ = false;
            out = constantBearingController(p_u_t + hoverVectTop, v_n_t, v_u_max_, dp_);
            dynhoverHeight_ = hoverHeight_;
            break;
        }

        case mode::lower : {
            enableLand_ = false;
            if (dynhoverHeight_ > posTolLowerHeight_ / 2) dynhoverHeight_ = dynhoverHeight_ - decreaseRate_ * dt;
            else dynhoverHeight_ = posTolLowerHeight_ / 2;
            hoverVect_ << 0, 0, -dynhoverHeight_;
            out = constantBearingController(p_u_t + hoverVect_, v_n_t, v_u_max_, dp_);
            break;
        }

        case mode::gainAdjust : {
            enableLand_ = false;
            if (dynhoverHeight_ > posTolLowerHeight_ / 2) dynhoverHeight_ = dynhoverHeight_ - decreaseRate_ * dt;
            else dynhoverHeight_ = posTolLowerHeight_ / 2;
            hoverVect_ << 0, 0, -dynhoverHeight_;
            out = constantBearingController(p_u_t + hoverVect_, v_n_t, v_u_max_, dp_gainAdjust_);
            break;
        }

        case mode::finalStage:{
            enableLand_ = false;
            if (dynhoverHeight_ > posTolLowerHeight_ / 2) dynhoverHeight_ = dynhoverHeight_ - finalStageDecentVel_ * dt;
            else dynhoverHeight_ = posTolLowerHeight_ / 2;
            hoverVect_ << 0, 0, -dynhoverHeight_;
            out = constantBearingController(p_u_t + hoverVect_, v_n_t, v_u_max_, dp_gainAdjust_);
            break;
        }

        case mode::land :{}
        case mode::landed : {
            enableLand_ = true;
        }
        default: {
            out = Eigen::Vector3d::Zero(3, 1);
        }
    }
    return out;
}

void Autopilot::loadParameterFile() {
    v_u_max_=ReadRosParam::readParameterDouble("controlParameters/v_u_max");
    dp_=ReadRosParam::readParameterDouble("controlParameters/dp");
    dp_gainAdjust_=ReadRosParam::readParameterDouble("controlParameters/dp_gainAdjust");
    hoverHeight_=ReadRosParam::readParameterDouble("controlParameters/hoverHeight");
    decreaseRate_=ReadRosParam::readParameterDouble("controlParameters/decreaseRate");
    posTolIntercept_=ReadRosParam::readParameterDouble("controlParameters/posTolIntercept");
    covarTolhover_=ReadRosParam::readParameterDouble("controlParameters/covarTolhover");
    posTolLowerRadius_=ReadRosParam::readParameterDouble("controlParameters/posTolLowerRadius");
    posTolLowerHeight_=ReadRosParam::readParameterDouble("controlParameters/posTolLowerHeight");
    gainAdjustAlt_=ReadRosParam::readParameterDouble("controlParameters/gainAdjustAltitude");
    finalStageAlt_=ReadRosParam::readParameterDouble("controlParameters/finalStageAlt");
    finalStageDecentVel_=ReadRosParam::readParameterDouble("controlParameters/finalStageDecentVel");
}

bool Autopilot::getEnambleLand() {
    return enableLand_;
}

double Autopilot::getHoverHeight() {
    return hoverVect_.z();
}

Eigen::Vector3d Autopilot::constantBearingController(Eigen::Vector3d p_u_t, Eigen::Vector3d v_n_t,
                                                           double v_u_max, double dp)
{
    /* Constant Bearing Guidance controller
     * From page 244 in Handbook of Marine Craft Hydrodynamics and Motion Control
     * written by Thor I. Fossen
     *
     * p_u_t -> position vector from UAV to target
     * p_t_u -> position vector from target to UAV
     * v_n_u -> velocity vector of the UAV relative to NED
     * v_n_t -> velocity vector of the tag relative to NED
     *
     * Tested by Line 28.02.18
     */
    Eigen::Vector3d p_t_u=-p_u_t;

    double K = v_u_max*p_t_u.norm()/sqrt(p_t_u.transpose()*p_t_u+pow(dp,2.0));
    Eigen::Vector3d v_a = -K*p_t_u/(p_t_u.norm()+0.1);
    Eigen::Vector3d v_n_u = v_n_t + v_a;

    return v_n_u;
}