#include <ros/init.h>
#include <ros/node_handle.h>
#include <ros/param.h>
#include <mavros_msgs/GlobalPositionTarget.h>
#include <sensor_msgs/NavSatFix.h>
#include <sensor_msgs/Imu.h>
#include <geometry_msgs/PoseWithCovarianceStamped.h>
#include "kalmanFilter.h"
#include "coordinateFrames.h"
#include "uavCallback.h"
#include "landingPadCallback.h"
#include "arucoCallback.h"
#include "writeToFile.h"
#include <utils/readRosParam.h>
#include <node_msgs/navResults.h>
#include "navigationCallback.h"

int main(int argc, char** argv){

    // Set system parameters
    int n = 9;     // Number of states
    int m = 9;     // Number of measurements
    int o = 3;     // Number of inputs

    // Initializing ROS
    ros::init(argc, argv, "node_navigation");

    // Creates a node handle
    ros::NodeHandle nH;
    ROS_INFO("NAVIGATION: Starting up navigation node...");

    // Run the node at 5Hz
    double loopRate;
    ros::param::param<double>("navigation_loop_rate", loopRate, 10.0);
    ROS_INFO("NAVIGATION: Using loop rate of %f Hz", loopRate);

    ros::Rate rate(loopRate);

    //Read data from system parameter file
    bool landingPadStatic=ReadRosParam::readParameterBool("systemParameters/landingPadStatic");
    bool autoInitKf=ReadRosParam::readParameterBool("systemParameters/autoInitFilter");
    //bool requireCameraInput=ReadRosParam::readParameterBool("systemParameters/requireCameraInput");
    bool saveToFile = ReadRosParam::readParameterBool("systemParameters/saveToFIle");

    // global NED frame                                                                                 
    CoordinateFrames globalNED;

    // Kalman filter object
    kalmanFilter kf(m,n,o,autoInitKf);
    kf.setTargetDynamics(landingPadStatic);
    kf.readParametersFromFile();

    // Subscribe to the /mavros/state topic
    UavCallback uavCallback(globalNED,kf);
    ROS_INFO("NAVIGATION: Expecting system state on topic: /mavros/state" );
    ros::Subscriber sub1 = nH.subscribe("mavros/state",10,&UavCallback::stateCallback, &uavCallback);
    // Wait for UAV to be armed
    ROS_INFO("NAVIGATION: Waiting for UAV to arm");
    while(!uavCallback.isArmed())
    {
        ros::spinOnce();
        rate.sleep();
    }
    sub1.shutdown();

    //GNSS time_ref
    // Subscribes to the /mavros/time_reference
    ROS_INFO("NAVIGATION: Expecting time reference on topic: /mavros/time_reference" );
    sub1 = nH.subscribe("/mavros/time_reference",10,&UavCallback::UAV_timeCallback, &uavCallback);
    ROS_INFO("NAVIGATION: Waiting for first time reference");
    while (!uavCallback.hasDt()) {
        ros::spinOnce();
        rate.sleep();
    }
    ros::Duration dtGnssRos = uavCallback.get_GNSS_time()-ros::Time::now();
    sub1.shutdown();
    //ros::Duration dtGnssRos = uavCallback.get_dtGnssRos();
    ROS_INFO("NAVIGATION: Delta time reference is set to %i s and %i ns", dtGnssRos.sec, dtGnssRos.nsec);

    /*Data from UAV:*/
    // Subscribes to the /mavros/global_position/global topic
    sub1 = nH.subscribe("/mavros/global_position/global", 10, &UavCallback::UAVGlobalPosCallback, &uavCallback);

    // Subscribes to the  /mavros/imu/data topic
    ros::Subscriber sub2 = nH.subscribe("/mavros/imu/data", 10, &UavCallback::UAVOrientCallback, &uavCallback);

    // Subscribe to the /mavros/local_position/velocity
    ros::Subscriber sub3 = nH.subscribe("/mavros/local_position/velocity", 10, &UavCallback::UAVVelocityCallback, &uavCallback);

    /*Data from landing pad*/
    LandingPadCallback lpCallback(globalNED, kf);
    lpCallback.setDtGnssRos(dtGnssRos);
    // Subscribe to the /odometry/target topic from LP
    ros::Subscriber sub5 = nH.subscribe("/odometry/target",10,&LandingPadCallback::lpPoseVelCallbackControlTower, &lpCallback);

    //Wait for Kalman filter states to be initialized
    ROS_INFO("NAVIGATION: Waiting for Kalman filter states to be initialized");
    while(!kf.getStatesInitialized())
    {
        ros::spinOnce();
        rate.sleep();
    }
    ros::Time time_now=ros::Time::now();
    kf.setInitTime(time_now+dtGnssRos);

    /*Data from aruco tag detection*/
    ArucoCallback arCallback(kf);
    arCallback.setDtGnssRos(dtGnssRos);
    ros::Subscriber sub7 = nH.subscribe("/arucoNode/data",10,&ArucoCallback::poseArucoCallback, &arCallback);

    // Publisher
    //ros::Publisher pub = nH.advertise<node_msgs::navResults>("/navigationNode/stateEstimator",100);
    //ros::Publisher pub = nH.advertise<node_msgs::navResults>("/navigationNode/stateEstimator2",100);
    node_msgs::navResults navMsg;

    //Kalman filter data from ROS bag
    NavigationCallback navCallback;
    ros::Subscriber navSub = nH.subscribe("/navigationNode/stateEstimator",10,&NavigationCallback::statesCallback, &navCallback);
    ROS_INFO("CONTROLLER: Expecting estimates states on topic: /navigationNode/stateEstimator" );

    // Status from the controller node
    ros::Subscriber controller_sub = nH.subscribe("/cmd/velocity",10,&NavigationCallback::smCallback, &navCallback);

    // print data to file from
    WriteToFile wf;
    //std::ofstream myfile;
    //    myfile.open("KfStates.csv");
    //    wf.writeHeader(myfile,"KF_Time;KF_States;KF_Covariance");
    //}

    Eigen::MatrixXd estStates;
    Eigen::MatrixXd estCovar;
    boost::array<double,3> tempA;
    boost::array<double,9> tempB;


    while (ros::ok() ) {
        ros::spinOnce();

        time_now = ros::Time::now();
        navMsg.header.stamp=time_now+dtGnssRos;
        navMsg.header.frame_id="/NED";

        estStates=kf.getEstStates(time_now+dtGnssRos);
        estCovar=kf.getEstCov();

        tempA = {estStates(0,0),estStates(1,0),estStates(2,0)};
        navMsg.relative_pos=tempA;

        tempB = {estCovar(0,0),estCovar(0,1),estCovar(0,2),estCovar(1,0),estCovar(1,1),estCovar(1,2),
                 estCovar(2,0),estCovar(2,1),estCovar(2,2)};
        navMsg.relative_pos_covar=tempB;

        tempA = {estStates(3,0),estStates(4,0),estStates(5,0)};
        navMsg.vel_LP =tempA;

        tempB = {estCovar(3,3),estCovar(3,4),estCovar(3,5),estCovar(4,3),estCovar(4,4),estCovar(4,5),
                 estCovar(5,3),estCovar(5,4),estCovar(5,5)};
        navMsg.vel_LP_covar=tempB;


        tempA = {estStates(6,0),estStates(7,0),estStates(8,0)};
        navMsg.bias_gnss=tempA;

        tempB = {estCovar(6,6),estCovar(6,7),estCovar(6,8),estCovar(7,6),estCovar(7,7),estCovar(7,8),
                 estCovar(8,6),estCovar(8,7),estCovar(8,8)};
        navMsg.bias_gnss_covar=tempB;

        //pub.publish(navMsg);


        if(saveToFile)
        {
            //std::list<Eigen::MatrixXd> data = {estStates, estCovar.diagonal()};
            //wf.writeDataToFile(myfile,navMsg.header.stamp,data);
        }

        rate.sleep();

    }

    //myfile.close();

    return 0;
}