//
// Created by line on 25.01.18.
//
#include <node_navigation/tools/kalmanFilter.h>

kalmanFilter::kalmanFilter() {
    //kalmanFilter::updatePhiDelta(0.0);
}

kalmanFilter::kalmanFilter(int m, int n, int o, bool autoInitKf)
        : m_{m},n_{n},o_{o},I{Eigen::MatrixXd::Identity(n,n)}, z(m,1), H(m,n), R(m,m), Phi(n,n),
          Delta(n,o), Q(n,n), K(n,m),P(n,n),P_pri(n,n),x_hat(n,1), x_hat_pri(n,1), u_(o,1) {
    x_hat_pri.setZero(n,1);
    u_.setZero(o,1);
    Phi.setZero(n,n);
    autoInitKf_=autoInitKf;
}

void kalmanFilter::setInitStates(const Eigen::MatrixXd &x0_) {
    x_hat=x0_;
    x_hat_pri=x0_;
    setStatesInitialized();
    ROS_INFO("NAVIGATION: Kalman filter states initialized as:");
    std::cout << x_hat << std::endl;
}

void kalmanFilter::setStatesInitialized() {
    statesInitialized=true;
}

bool kalmanFilter::getStatesInitialized() {
    return statesInitialized;
}

bool kalmanFilter::getAutoInit() {
    return autoInitKf_;
}

void kalmanFilter::setInitTime(const ros::Time& timeNow_) {
    prevTime=timeNow_;
}

void kalmanFilter::projectAhead(){
    // Project ahead:
    x_hat_pri=Phi*x_hat+Delta*u_;
    P_pri=Phi*P*Phi.transpose()+Q;
}

void kalmanFilter::updateMeasurement(int type, Eigen::Vector3d& measurement, ros::Time& now) {
    kalmanFilter::update_H(type);
    kalmanFilter::update_z(type,measurement);

    if(check_dt(now))
    {//New time step k
        updatePhiDelta(dt);
        projectAhead();
    }

    if(getStatesInitialized())kalmanFilter::updateFilter();
}

void kalmanFilter::updateMeasurement(int type, Eigen::Vector3d& measurement, ros::Time& now, Eigen::Matrix3d& covariance) {
    kalmanFilter::update_R(type,covariance);
    kalmanFilter::updateMeasurement(type,measurement,now);
}

void kalmanFilter::updateFilter(){
    // Compute Kalman gain:
    K=P_pri*H.transpose()*(H*P_pri*H.transpose()+R).inverse();

    //Update estimate with measurement z:
    x_hat=x_hat_pri+K*(z-H*x_hat_pri);
    P=(I-K*H)*P_pri*(I-K*H).transpose()+K*R*K.transpose();

    // Set the predicted a priori in case of several measurement updates
    x_hat_pri=x_hat;
    P_pri=P;

    //print *x_hat_pri.data()@12

    //std::cout << "Estimated state:" << std::endl << x_hat << std::endl;
    //std::cout << "State covariance:" << std::endl << P << std::endl;
}

void kalmanFilter::setTargetDynamics(bool targetStatic) {
    staticTarget_=targetStatic;
}

void kalmanFilter::updatePhiDelta(double dt){
    //Reduced nr of states. Only position, linear velocity and GNSS bias
    Phi.setIdentity(n_,n_);
    Phi.block(0,3,3,3)=Eigen::MatrixXd::Identity(3,3)*dt;
    if(staticTarget_){
        Phi.block(3,3,3,3)=Eigen::MatrixXd::Zero(3,3);
    }

    Delta.setZero(n_,o_);
    Delta.topLeftCorner(3,3)=-Eigen::MatrixXd::Identity(3,3)*dt;

    // Debuging
    //std::cout << "Phi:" << std::endl << Phi << std::endl;
    //std::cout << "Delta:" << std::endl << Delta << std::endl;
}

void kalmanFilter::update_z(int type, Eigen::Vector3d measurement) {
    Eigen::VectorXd z_ = Eigen::VectorXd::Zero(m_);
    switch (type) {
        case kalmanFilter::measurementType::lP_pos :
            z_.segment(0,3)=measurement; break;

        case kalmanFilter::measurementType::arUco_pos :
            z_.segment(3,3)=measurement; break;

        case kalmanFilter::measurementType::lP_vel :
            z_.segment(6,3)=measurement; break;
    }
    z=z_;
}

void kalmanFilter::update_H(int type) {
    Eigen::MatrixXd H_ = Eigen::MatrixXd::Zero(m_,n_);
    Eigen::Matrix3d I_ = Eigen::Matrix3d::Identity(3,3);
    Eigen::Matrix3d O_ = Eigen::Matrix3d::Zero(3,3);
    switch (type) {
        case kalmanFilter::measurementType::lP_pos :
            H_.block(0,0,3,n_) << I_, O_, I_; break;
        case kalmanFilter::measurementType::arUco_pos :
            H_.block(3,0,3,n_) << I_, O_, O_; break;
        case kalmanFilter::measurementType::lP_vel :
            H_.block(6,0,3,n_) << O_, I_, O_; break;
    }
    H=H_;
}

void kalmanFilter::update_R(int type, Eigen::Matrix3d covarianse) {
    Eigen::MatrixXd R_ = R;
    switch (type) {
        case kalmanFilter::measurementType::lP_pos :
            R_.block(0,0,3,3)=covarianse; break;
        case kalmanFilter::measurementType::arUco_pos :
            R_.block(3,3,3,3)=covarianse; break;
        case kalmanFilter::measurementType::lP_vel :
            R_.block(6,6,3,3)=covarianse; break;
    }
    R=R_;
}

void kalmanFilter::set_u(const Eigen::Vector3d &u) {
    u_=u;
}

void kalmanFilter::set_UAV_orient(const Eigen::Quaterniond uav_orient, const ros::Time &timeStamp,
                                  const Eigen::Vector3d& covariance) {
    kalmanFilter::bufferElement temp_element;
    temp_element.timestamp=timeStamp;
    temp_element.element=uav_orient.coeffs();
    temp_element.covariance=covariance;

    kalmanFilter::addElementToBuffer(uav_orient_,20,temp_element);
}

kalmanFilter::bufferElement kalmanFilter::get_UAV_orient(const ros::Time &timestamp) {
    return kalmanFilter::findCorrespondingElement(timestamp,uav_orient_);
}

bool kalmanFilter::uav_position_available(){
    bool tempSize = uav_position_.size()>0;
    return tempSize;
}

bool kalmanFilter::uav_orient_available() {
    bool tempSize = uav_orient_.size()>0;
    return tempSize;
}

void kalmanFilter::set_UAV_position(const Eigen::Vector3d uav_position, const ros::Time &timeStamp,
                                    const Eigen::Vector3d& covariance) {
    kalmanFilter::bufferElement temp_element;
    temp_element.timestamp=timeStamp;
    temp_element.element=uav_position;
    temp_element.covariance=covariance;

    kalmanFilter::addElementToBuffer(uav_position_,20,temp_element);
}

kalmanFilter::bufferElement kalmanFilter::get_UAV_position(const ros::Time &timestamp) {
    return kalmanFilter::findCorrespondingElement(timestamp,uav_position_);
}

bool kalmanFilter::check_dt(ros::Time timeNow) {
    bool tempOut=false;
    double min_dt = 0.005;
    if((timeNow - prevTime ).toSec()>min_dt)
    {
        dt = (timeNow - prevTime ).toSec();
        prevTime=timeNow;
        timeInitialized=true;
        tempOut=true;

        //std::cout << "dt is set to: " << dt << std::endl;
    }
    return tempOut;
}

void kalmanFilter::deadReckoning(ros::Time& timeNow) {
    if(check_dt(timeNow))
    {//New time step k
        updatePhiDelta(dt);
        projectAhead();

        x_hat=x_hat_pri;
        P=P_pri;
    }
}

Eigen::MatrixXd kalmanFilter::getEstStates(ros::Time tNow){
    kalmanFilter::deadReckoning(tNow);

    return x_hat;
}

Eigen::MatrixXd kalmanFilter::getEstCov(){
    return P;
}

bool kalmanFilter::readParametersFromFile() {
    /*
    * 	Read parameters from .yaml files
    */
    bool result = true;
    struct matrices{
        std::string name;
        Eigen::MatrixXd& matrix;
    };

    std::list<matrices> mat_to_read;
    mat_to_read.push_front(matrices{"Q",Q});
    mat_to_read.push_front(matrices{"R",R});
    mat_to_read.push_front(matrices{"x0",x0});
    mat_to_read.push_front(matrices{"P0",P0});

    std::vector<float> tempVector;
    ros::NodeHandle parameterNode;
    int rows, cols;

    for (std::list<matrices>::iterator it=mat_to_read.begin(); it != mat_to_read.end(); ++it){
        if(parameterNode.getParam("/filterParameters/"+it->name+"/data",tempVector))
        {
            parameterNode.getParam("/filterParameters/"+it->name+"/rows",rows);
            parameterNode.getParam("/filterParameters/"+it->name+"/cols",cols);
            it->matrix.resize(rows,cols);
            for(int i=0;i<rows*cols;i++)
            {
                it->matrix(int(floor(i/cols)),i%cols)=tempVector[i];
            }
        }
        else
        {
            ROS_WARN("Error loading Kalman filter parameters!");
            result=false;
        }
    }

    if(result)
    {
        if(!autoInitKf_) setInitStates(x0);
        P=P0;
        P_pri=P0;
    }
    //std::cout << "Q:" << std::endl << Q << std::endl;
    //std::cout << "R:" << std::endl << R << std::endl;
    //std::cout << "x0:" << std::endl << x0 << std::endl;
    //std::cout << "P0:" << std::endl << P0 << std::endl;
    return result;
}

void kalmanFilter::addElementToBuffer(std::list<bufferElement> &buffer, unsigned long bufferSize,
                                      const bufferElement &element) {
    while(buffer.size()>bufferSize-1)
    {
        buffer.pop_back();
    }
    buffer.push_front(element);
}

kalmanFilter::bufferElement kalmanFilter::findCorrespondingElement(const ros::Time &timestamp,
                                                     std::list<kalmanFilter::bufferElement> &buffer) {
    //This function finds the measurement closest to the input timestamp and gives an output of the interpolation
    // between the closest time stemps and the required time step.

    //Find nearest value
    auto it = buffer.begin();
    while(it != buffer.end() && it->timestamp>timestamp)
    {
        it++;
    }

    kalmanFilter::bufferElement ret_element;
    //If end of list
    if(it==buffer.end())
    {
        //Return last
        ret_element=*std::prev(it);
    }
    else if(it==buffer.begin()) //First element
    {
        //Return first
        ret_element=*it;
    }
    else
    {
        Eigen::VectorXd dudt=(std::prev(it)->element-it->element)/(std::prev(it)->timestamp.toSec()-it->timestamp.toSec());

        ret_element.element=dudt*timestamp.toSec()+std::prev(it)->element-std::prev(it)->timestamp.toSec()*dudt;
        ret_element.timestamp=timestamp;
        ret_element.covariance=it->covariance;
    }

    return ret_element;
}

void kalmanFilter::testFunctions(){
    std::list<kalmanFilter::bufferElement> testList;

    std::vector<Eigen::Vector3d> data_vector;
    data_vector.emplace_back(Eigen::Vector3d(5,6,7));
    data_vector.emplace_back(Eigen::Vector3d(4,5,6));
    data_vector.emplace_back(Eigen::Vector3d(3,4,5));
    data_vector.emplace_back(Eigen::Vector3d(2,3,4));
    data_vector.emplace_back(Eigen::Vector3d(1,2,3));

    std::vector<ros::Time> time_vector;
    time_vector.emplace_back(ros::Time(6.0));
    time_vector.emplace_back(ros::Time(5.0));
    time_vector.emplace_back(ros::Time(4.0));
    time_vector.emplace_back(ros::Time(3.0));
    time_vector.emplace_back(ros::Time(2.0));

    kalmanFilter::bufferElement temp_buffer;
    for (int i = 0; i < 5; ++i) {
        temp_buffer.timestamp=time_vector.back();
        temp_buffer.element=data_vector.back();
        time_vector.pop_back();
        data_vector.pop_back();

        kalmanFilter::addElementToBuffer(testList,3,temp_buffer);
    }

    //Eigen::Vector3d ret_vect = kalmanFilter::findCorrespondingElement(ros::Time(4.5),testList);

    //std::cout << "Vector: " << ret_vect << std::endl;
}