//
// Created by Line on 20.03.18.
//

#include "node_navigation/tools/landingPadCallback.h"

LandingPadCallback::LandingPadCallback(CoordinateFrames& NED, kalmanFilter& kf_)
        :globalNED{NED},kf{kf_}{
    //Read parameters from parameter file
    //Compensate for the distance between the LP body frame and ArUco tag
    p_lb_l=ReadRosParam::readParameterMatrix("systemParameters/translationTagBodyLP");
    //Rotation from landing pad to body
    R_lb=ReadRosParam::readParameterMatrix("systemParameters/rotationTagBodyLP");
    enableWriteToFile = ReadRosParam::readParameterBool("systemParameters/saveToFIle");

    if(enableWriteToFile){
        myfile.open("KfLandingPad.csv");
        wf.writeHeader(myfile,"KF_LP_Time;KF_LP_vel_NED;KF_rel_pos_NED");
    }
}

//Reading data form the Landing Pad
void LandingPadCallback::lpPoseVelCallbackControlTower(const nav_msgs::Odometry::Ptr& lpPoseVel)
{
    //Check if NED is defined
    if(globalNED.isNedDefined())
    {
        // Read out the lat long and alt data
        CoordinateFrames::geodetic lpGeodetic;
        lpGeodetic.longitude=lpPoseVel->pose.pose.position.y*180.0/M_PI;    //Convert from radians to deg
        lpGeodetic.latitude=lpPoseVel->pose.pose.position.x*180.0/M_PI;
        lpGeodetic.altitude=lpPoseVel->pose.pose.position.z;

        // Read out the orientation and convert to euler angles
        // Get orientation data from the Landing pad (quaternions)
        Eigen::Quaterniond orient_ENU(lpPoseVel->pose.pose.orientation.w,lpPoseVel->pose.pose.orientation.x,lpPoseVel->pose.pose.orientation.y,lpPoseVel->pose.pose.orientation.z);
        lp_pose.orientation=CoordinateFrames::ENU2NED_quat(orient_ENU);

        // Read out euler angles
        //Eigen::Vector3d orient_NED_euler = orient_NED.toRotationMatrix().eulerAngles(0,1,2);

        // Read out the pose covariance matrix
        lp_pose.covariance=Eigen::Map<Eigen::MatrixXd>(lpPoseVel->pose.covariance.data(),6,6);

        // Transform lat long coordinates to the global NED frame
        lp_pose.position=globalNED.getVectNED2geodetic(lpGeodetic);

        // Todo: Add clock calibration to the control tower ROS node!
        //ros::Duration dtUavGnss(66001341,8789351);
        // Read out the timing given in GNSS time
        //lp_pose.timestamp=lpPoseVel->header.stamp+dtUavGnss;
        //lp_vel.timestamp=lpPoseVel->header.stamp+dtUavGnss;

        //lp_pose.timestamp=lpPoseVel->header.stamp+dtGnssRos;
        lp_pose.timestamp=lpPoseVel->header.stamp;
        lp_vel.timestamp=lp_pose.timestamp;

        //Read out the linear and angular velocity
        Eigen::Vector3d lp_vel_linear_ENU=Eigen::Vector3d(lpPoseVel->twist.twist.linear.x,lpPoseVel->twist.twist.linear.y,lpPoseVel->twist.twist.linear.z);
        lp_vel.linear=CoordinateFrames::ENU2NED_euler(lp_vel_linear_ENU);
        lp_vel.angular=Eigen::Vector3d(lpPoseVel->twist.twist.angular.x,lpPoseVel->twist.twist.angular.y,lpPoseVel->twist.twist.angular.z);

        // Read out the velocity covariance matrix
        lp_vel.covariance=Eigen::Map<Eigen::MatrixXd>(lpPoseVel->twist.covariance.data(),6,6);
        Eigen::Matrix3d lp_vel_covar_linear = lp_vel.covariance.block(0,0,3,3);

        // Update kalman filter // Todo: Do not include covariance matrix if static
        kf.updateMeasurement(kalmanFilter::measurementType::lP_vel,lp_vel.linear,lp_vel.timestamp,lp_vel_covar_linear);

        if(kf.uav_position_available())
        {
            //Finds the corresponding measurement from UAV_position that matches the time stamp
            kalmanFilter::bufferElement uav_pos = kf.get_UAV_position(lp_pose.timestamp);
            Eigen::Vector3d uav_pos_NED=uav_pos.element;


            //Rotation from body to NED
            Eigen::Matrix3d R_bn = lp_pose.orientation.toRotationMatrix();

            // LP position given in NED
            Eigen::Vector3d lp_pos_NED=lp_pose.position-R_bn*R_lb*p_lb_l;

            //Adds the covariance matrix for UAV and LP position
            Eigen::Matrix3d lp_uav_covar = Eigen::Matrix3d(lp_pose.covariance.topLeftCorner(3,3)) + Eigen::Matrix3d(uav_pos.covariance.asDiagonal());

            //IF first measurement, set it as initial conditions
            if (!kf.getStatesInitialized()&&kf.getAutoInit())
            {
                Eigen::MatrixXd temp_x(9,1);
                temp_x.block(0,0,3,1)=Eigen::MatrixXd::Zero(3,1);
                temp_x.block(3,0,3,1)=lp_vel.linear;
                temp_x.block(6,0,3,1)=lp_pos_NED-uav_pos_NED;

                kf.setInitStates(temp_x);
            }

            //Update the Kalman filter
            Eigen::Vector3d rel_pos_NED = lp_pos_NED - uav_pos_NED;
            kf.updateMeasurement(kalmanFilter::measurementType::lP_pos, rel_pos_NED, lp_pose.timestamp,
                                 lp_uav_covar);

            //Write to file
            std::list<Eigen::MatrixXd> my_data = {lp_vel.linear, rel_pos_NED};
            if (enableWriteToFile)wf.writeDataToFile(myfile, lp_pose.timestamp, my_data);

        }
    }
    else
    {
        ROS_INFO("NAVIGATION: Global NED is not defined");
    }
}

LandingPadCallback::pose LandingPadCallback::get_pose(){
    return lp_pose;
}

LandingPadCallback::velocity LandingPadCallback::get_velocity(){
    return lp_vel;
}

void LandingPadCallback::setDtGnssRos(ros::Duration dtGnssRos_) {
    dtGnssRos=dtGnssRos_;
}
